"""
Test transport function for `in3.client.Client`.

Collect mocked responses from `http.py`.
"""
import hashlib

from in3.libin3.transport import In3Request, In3Response
from tests.network_error.mock import http


def mock_transport(in3_request: In3Request, in3_response: In3Response):
    """
    Transports each request coming from libin3 to the in3 network and and reports the answer back
    Args:
        in3_request (In3Request): request sent by the In3 Client Core to the In3 Network
        in3_request (In3Response): response to be dispatched to the In3 Client Core
    Returns:
        exit_status (int): Always zero for signaling libin3 the function executed OK.
    """

    for i in range(0, in3_request.urls_len()):
        try:
            http_request = {
                'url': in3_request.url_at(i),
                'data': in3_request.payload(),
                'headers': {'Content-type': 'application/json'},
                'timeout': in3_request.timeout()
            }
            uuid = hashlib.sha1()
            uuid.update(http_request['url'])
            uuid.update(http_request['data'])
            response = http.data[http_request['url']][uuid]
            in3_response.success(i, response)
        except Exception as err:
            err_bytes = str(err).encode('utf8')
            in3_response.failure(i, err_bytes)
    return 0



"""
Load libin3 shared library for the current system, map function ABI, sets in3 network transport functions.
"""
import ctypes as c
import platform

from pathlib import Path

DEBUG = False


def _load_shared_library():
    """
    Loads library depending on host system.
    """

    def platform_selector(prefix: str, lib_path: str) -> str:
        system, node, release, version, machine, processor = platform.uname()

        global DEBUG
        suffix = None
        if processor in ('i386', 'x86_64') or 'Intel' in processor or 'AMD' in processor:
            if '64' in machine:
                suffix = 'x64'
            elif '32' in machine or '86' in machine:
                suffix = 'x86'
            elif 'arm' in machine.lower():
                suffix = 'arm7'
            if DEBUG:
                suffix += 'd'
            if system == 'Windows':
                suffix += ".dll"
            elif system == "Linux":
                suffix += ".so"
            elif system == 'Darwin':
                suffix += ".dylib"
        if not suffix:
            raise OSError()
        return str(Path(lib_path, "{}.{}".format(prefix, suffix)))

    def fallback_platform_selector(search_string: str):
        import glob

        system, node, release, version, machine, processor = platform.uname()
        lib_list = glob.glob(search_string)
        for lib in lib_list:
            try:
                lib_instance = c.cdll.LoadLibrary(lib)
                return lib_instance
            except Exception:
                pass
        raise OSError('Not available on this platform ({}, {}, {}).'.format(system, processor, machine))

    path = Path(Path(__file__).parent, "shared")
    try:
        return c.cdll.LoadLibrary(platform_selector('libin3', path))
    except OSError:
        return fallback_platform_selector(str(path) + '/*')


_libin3 = _load_shared_library()


def libin3_new(chain_id: int, transport_fn: c.CFUNCTYPE) -> int:
    """
    Instantiate new In3 Client instance.
    Args:
        chain_id (int): Chain id as integer
        transport_fn: Transport function for the in3 network requests
        storage_fn: Cache Storage function for node list and requests caching
    Returns:
         instance (int): Memory address of the client instance, return value from libin3_new
    """

    def map_function_signatures():
        # =================== LIBIN3 SHARED LIBRARY MAPPING ===================
        _libin3.in3_for_chain_auto_init.argtypes = c.c_int,
        _libin3.in3_for_chain_auto_init.restype = c.c_void_p
        _libin3.in3_free.argtypes = c.c_void_p,
        _libin3.eth_set_pk_signer_hex.argtypes = c.c_void_p, c.c_char_p
        _libin3.in3_client_rpc.argtypes = c.c_void_p, c.c_char_p, c.c_char_p, c.POINTER(c.c_char_p), c.POINTER(
            c.c_char_p)
        _libin3.in3_client_rpc.restype = c.c_int
        _libin3.in3_req_add_response.argtypes = c.c_void_p, c.c_int, c.c_bool, c.c_char_p, c.c_int

    assert isinstance(chain_id, int)
    map_function_signatures()
    _libin3.in3_set_default_transport(transport_fn)
    # TODO: in3_set_default_signer
    _libin3.in3_register_eth_full()
    # TODO: IPFS libin3.in3_register_ipfs();
    _libin3.in3_register_eth_api()
    global DEBUG
    if DEBUG:
        # set logger level to TRACE
        _libin3.in3_log_set_quiet_(False)
        _libin3.in3_log_set_level_(0)
    return _libin3.in3_for_chain_auto_init(chain_id)


def libin3_free(instance: int):
    """
    Free In3 Client objects from memory.
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
    """
    _libin3.in3_free(instance)


def libin3_call(instance: int, fn_name: bytes, fn_args: bytes) -> (str, str):
    """
    Make Remote Procedure Call to an arbitrary method of a libin3 instance
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
        fn_name (bytes): Name of function that will be called in the client rpc.
        fn_args: (bytes) Serialized list of arguments, matching the parameters order of this function. i.e. ['0x123']
    Returns:
        result (int): Function execution status.
    """
    response = c.c_char_p()
    error = c.c_char_p()
    result = _libin3.in3_client_rpc(instance, fn_name, fn_args, c.byref(response), c.byref(error))
    return result, response.value, error.value


def libin3_set_pk(instance: int, private_key: bytes):
    """
    Register the signer module in the In3 Client instance, with selected private key loaded in memory.
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
        private_key: 256 bit number.
    """
    _libin3.eth_set_pk_signer_hex(instance, private_key)


def libin3_in3_req_add_response(*args):
    """
    Transport function that registers a response to a request.
    Args:
        *args:
    """
    _libin3.in3_req_add_response(*args)



"""
Integrated negative tests for `in3.eth` module.
"""
import unittest

import in3
from tests.integrated.mock.config import mock_config
from tests.integrated.mock.transport import mock_transport


class EthereumNegativeTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_get_block_by_number_client(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.eth.block_by_number(9937219)
            with self.assertRaises(in3.ClientException):
                self.client.eth.block_by_number(True)

    # def test_get_block_by_number_client_issue_334(self):
    #     for i in range(20):
    #         with self.assertRaises(in3.ClientException):
    #             self.client.eth.block_by_number(False)

    def test_get_transaction_by_hash_client(self):
        tx_bad_hash = '0xTe25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668aeP'
        tx_wrong_hash = '0xTe25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668aeb'
        tx_evil_hash = '0xa™a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668aeP'
        for i in range(20):
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_by_hash(tx_bad_hash)
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_by_hash(tx_wrong_hash)
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_by_hash(tx_evil_hash)

    def test_get_transaction_receipt_client(self):
        tx_bad_hash = '0xb13b9d38642216af2545f1b9f882413bcdef13bec21def57c699d3a967d763bP'
        tx_wrong_hash = '0xb13b9d38642216af2545f1b9f882413bcdef13bec21def57c699d3a967d763bb'
        tx_evil_hash = '0xa™a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668aeP'
        for i in range(20):
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_receipt(tx_bad_hash)
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_receipt(tx_wrong_hash)
            with self.assertRaises(in3.ClientException):
                self.client.eth.transaction_receipt(tx_evil_hash)


class ParsingTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_get_block_by_number(self):
        with self.assertRaises(AssertionError):
            self.client.eth.block_by_number(None)
        with self.assertRaises(AssertionError):
            self.client.eth.block_by_number(-1)
        with self.assertRaises(AssertionError):
            self.client.eth.block_by_number('œ∑´´†√¨')

    def test_get_transaction_by_hash(self):
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(None)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(-1)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash('œ∑´´†√¨')
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(True)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(False)
        tx_hash = '0xae25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668ae'
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(tx_hash)
        tx_hash = '0xae25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668ae77'
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_by_hash(tx_hash)

    def test_get_tx_receipt(self):
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(None)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(-1)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt('œ∑´´†√¨')
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(True)
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(False)
        tx_hash = '0xae25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668ae'
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(tx_hash)
        tx_hash = '0xae25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668ae77'
        with self.assertRaises(in3.HashFormatException):
            self.client.eth.transaction_receipt(tx_hash)


class NegativeGoerliTest(EthereumNegativeTest):

    def setUp(self):
        # self.client = in3.Client('goerli', in3_config=mock_config)
        self.client = in3.Client('goerli', in3_config=mock_config, transport=mock_transport)


class NegativeKovanTest(EthereumNegativeTest):

    def setUp(self):
        # self.client = in3.Client('kovan', in3_config=mock_config)
        self.client = in3.Client('kovan', in3_config=mock_config, transport=mock_transport)


if __name__ == '__main__':
    unittest.main()



"""
Integrated negative tests for `in3` module. Doesnt test submodules.
"""
import unittest

import in3
from tests.integrated.mock.config import mock_config
from tests.integrated.mock.transport import mock_transport


class ClientNegativeTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_configure(self):
        # TODO
        pass

    def test_ens_owner(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_owner('depraz.eth', '')
            with self.assertRaises(in3.ClientException):
                self.client.ens_owner('deprazz.eth')

    def test_ens_resolver(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_resolver('depraz.eth', '')
            with self.assertRaises(in3.ClientException):
                self.client.ens_owner('deprazz.eth')

    def test_ens_address(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_address('depraz.eth', '')
            with self.assertRaises(in3.ClientException):
                self.client.ens_owner('deprazz.eth')

    def test_ens_namehash(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_namehash('0x0.eth')


class ClientParsingTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_instantiate(self):
        with self.assertRaises(AssertionError):
            in3.Client(None)
        with self.assertRaises(AssertionError):
            in3.Client(1)
        with self.assertRaises(AssertionError):
            in3.Client(-1)
        with self.assertRaises(AssertionError):
            in3.Client('œ∑´´†√¨')
        with self.assertRaises(AssertionError):
            in3.Client('!@# asd')
        with self.assertRaises(AssertionError):
            in3.Client({1: 1})
        with self.assertRaises(AssertionError):
            in3.Client((1))
        with self.assertRaises(AssertionError):
            in3.Client([1])

    def test_configure(self):
        with self.assertRaises(AssertionError):
            in3.Client(in3_config=1)
        with self.assertRaises(AssertionError):
            in3.Client(in3_config=-1)
        with self.assertRaises(AssertionError):
            in3.Client(in3_config='œ∑´´†√¨')
        with self.assertRaises(AssertionError):
            in3.Client(in3_config={1: 1})
        with self.assertRaises(AssertionError):
            in3.Client(in3_config=(1))
        with self.assertRaises(AssertionError):
            in3.Client(in3_config=[1])

    def test_ens_owner(self):
        # Other calls like `addr` require more than one eth_call, being more complex to mock the tests. Suffice for now.
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', '0x')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', '¡™£¢∞§¶•')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', 132)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', True)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', '123')
        with self.assertRaises(AssertionError):
            self.client.ens_owner('depraz')
        with self.assertRaises(AssertionError):
            self.client.ens_owner('depraz.')
        with self.assertRaises(AssertionError):
            self.client.ens_owner('depraz.etho')
        with self.assertRaises(AssertionError):
            self.client.ens_owner('.eth')

    def test_ens_resolver(self):
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_resolver('depraz.eth', '0x')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_resolver('depraz.eth', '¡™£¢∞§¶•')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_resolver('depraz.eth', 132)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_resolver('depraz.eth', True)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_resolver('depraz.eth', '123')
        with self.assertRaises(AssertionError):
            self.client.ens_resolver('depraz')
        with self.assertRaises(AssertionError):
            self.client.ens_resolver('depraz.')
        with self.assertRaises(AssertionError):
            self.client.ens_resolver('depraz.etho')
        with self.assertRaises(AssertionError):
            self.client.ens_resolver('.eth')

    def test_ens_address(self):
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_address('depraz.eth', '0x')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_address('depraz.eth', '¡™£¢∞§¶•')
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_address('depraz.eth', 132)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_address('depraz.eth', True)
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_address('depraz.eth', '123')
        with self.assertRaises(AssertionError):
            self.client.ens_address('depraz')
        with self.assertRaises(AssertionError):
            self.client.ens_address('depraz.')
        with self.assertRaises(AssertionError):
            self.client.ens_address('depraz.etho')
        with self.assertRaises(AssertionError):
            self.client.ens_address('.eth')

    def test_ens_namehash(self):
        with self.assertRaises(AssertionError):
            self.client.ens_namehash('depraz')
        with self.assertRaises(AssertionError):
            self.client.ens_namehash('depraz.')
        with self.assertRaises(AssertionError):
            self.client.ens_namehash('depraz.etho')
        with self.assertRaises(AssertionError):
            self.client.ens_namehash('.eth')


class KovanClientTest(ClientNegativeTest):

    def setUp(self):
        # self.client = in3.Client('kovan', in3_config=mock_config)
        self.client = in3.Client('kovan', in3_config=mock_config, transport=mock_transport)

    def test_ens_resolver(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_resolver('depraz.eth')

    def test_ens_address(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_address('depraz.eth')

    def test_ens_owner(self):
        for i in range(50):
            with self.assertRaises(in3.ClientException):
                self.client.ens_owner('depraz.eth')


class GoerliClientTest(ClientNegativeTest):

    def setUp(self):
        # self.client = in3.Client('goerli', in3_config=mock_config)
        self.client = in3.Client('goerli', in3_config=mock_config, transport=mock_transport)


if __name__ == '__main__':
    unittest.main()



import ctypes as c

from in3.libin3.rpc_api import libin3_in3_req_add_response


class NativeRequest(c.Structure):
    """
    Based on in3/client/.h in3_request_t struct
    """
    _fields_ = [("payload", c.POINTER(c.c_char)),
                ("urls", c.POINTER(c.POINTER(c.c_char))),
                ("urls_len", c.c_int),
                ("results", c.c_void_p),
                ("timeout", c.c_uint32),
                ("times", c.c_uint32)]


class NativeResponse(c.Structure):
    """
    Based on in3/client/.h in3_response_t struct
    """
    """
    typedef struct in3_response {
      sb_t error;  /**< a stringbuilder to add any errors! */
      sb_t result; /**< a stringbuilder to add the result */
    """


class In3Request:
    """
    C level abstraction of an Incubed request.
    """

    def __init__(self, in3_request: NativeRequest):
        self.in3_request = in3_request

    def url_at(self, index: int):
        """
        Gets the `index` url on the request url node list.
        Args:
            index (int): Positional argument to retrieve the url of a node from the list of urls. The total length should be retreived with `urls_len`
        Returns:
            fn_return (str): The url of a node to request a response from.
        """
        return c.string_at(self.in3_request.contents.urls[index])

    def urls_len(self):
        """
        Gets the the size of the request url node list
        """
        return self.in3_request.contents.urls_len

    def payload(self):
        """
        Gets the payload to be sent
        """
        return c.string_at(self.in3_request.contents.payload)

    def timeout(self):
        """
        Get timeout of the request, `0` being no set timeout
        """
        return self.in3_request.contents.timeout


class In3Response:
    """
    C level abstraction of an Incubed response.
    """

    def __init__(self, in3_response: NativeResponse):
        self.in3_response = in3_response

    def success(self, index: int, msg: bytes):
        """
        Function to be invoked in order to write the result for the request in case of success
        Args:
            index (int): Positional argument related to which url on the `In3Request` list this response is associated with. Use `In3Request#url_at` to get the url. The value of both parameters are shared
            msg (str): The actual response to be returned to in3 client
        """
        libin3_in3_req_add_response(self.in3_response, index, False, msg, len(msg))

    def failure(self, index: int, msg: bytes):
        """
        Function to be invoked in order to write the result for the request in case of failure
        Args:
            index (int): Positional argument related to which url on the `In3Request` list this response is associated with. Use `In3Request#url_at` to get the url. The value of both parameters are shared.
            msg (str): The actual response to be returned to in3 client.
        """
        libin3_in3_req_add_response(self.in3_response, index, True, msg, len(msg))


# TODO: Move to a OO perspective
def factory(transport_fn):
    """
    C level abstraction of a transport handler.
    Decorates a transport function augmenting its capabilities for native interoperability
    """

    def new(native_request: NativeRequest):
        request = In3Request(native_request)
        response = In3Response(native_request)
        return transport_fn(request, response)

    # the transport function to be implemented by the transport provider.
    # typedef in3_ret_t (*in3_transport_send)(in3_request_t* request);
    c_transport_fn_interface = c.CFUNCTYPE(c.c_int, c.POINTER(NativeRequest))
    return c_transport_fn_interface(new)



"""
Encapsulates low-level rpc calls into a comprehensive runtime.
"""
import json
from enum import Enum

import in3.libin3.transport as transport
from in3.exception import ClientException
from in3.libin3.enum import RPCCode
from in3.libin3.rpc_api import libin3_new, libin3_free, libin3_call, libin3_set_pk


class RPCCallRequest:
    """
    RPC request to libin3
    Args:
        fn_name: Name of function that will be called in libin3
        fn_args: Arguments matching the parameters order of this function
    """

    def __init__(self, fn_name: str or Enum, fn_args: tuple = None, formatted: bool = False):
        self.fn_name = str(fn_name).encode('utf8')
        if fn_args and formatted:
            self.fn_args = fn_args[0].encode('utf8')
        elif fn_args and not formatted:
            self.fn_args = json.dumps(list(fn_args)).replace('\'', '').encode('utf8')
        else:
            self.fn_args = '[]'.encode('utf8')


class In3Runtime:
    """
    Instantiate libin3 and frees it when garbage collected.
    Args:
        chain_id (int): Chain-id based on EIP-155. If None provided, will connect to the Ethereum network. i.e: 0x1 for mainNet
        transport_fn: Transport function to handle the HTTP Incubed Network requests.
    """

    def __init__(self, chain_id: int, transport_fn):
        self.transport_handler = transport.factory(transport_fn)
        self.in3 = libin3_new(chain_id, self.transport_handler)
        self.chain_id = chain_id

    def __del__(self):
        # libin3_free(self.transport_handler)
        # libin3_free(self.storage_handler)
        libin3_free(self.in3)

    def call(self, fn_name: str or Enum, *fn_args, formatted: bool = False) -> str or dict:
        """
        Make a remote procedure call to a function in libin3
        Args:
            fn_name (str or Enum): Name of the function to be called
            fn_args: Arguments matching the parameters order of this function
            formatted (bool): True if args must be sent as-is to RPC endpoint
        Returns:
            fn_return (str): String of values returned by the function, if any.
        """
        request = RPCCallRequest(fn_name, fn_args, formatted)
        result, response, error = libin3_call(self.in3, request.fn_name, request.fn_args)
        in3_code = RPCCode(result)
        if not in3_code == RPCCode.IN3_OK or error:
            raise ClientException(str(error))
        return json.loads(response)

    # TODO: Refactor for the new signer api
    def set_signer_account(self, secret: int) -> int:
        """
        Load an account secret to sign Ethereum transactions with `eth_sendTransaction` and `eth_call`.
        Args:
            secret: SK 256 bit number. example: int(0x387a8233c96e1fc0ad5e284353276177af2186e7afa85296f106336e376669f7, 16)
        """
        return libin3_set_pk(self.in3, hex(secret).encode('utf8'))



import http.client
from http import HTTPStatus

from in3.libin3.transport import In3Request, In3Response


def http_transport(in3_request: In3Request, in3_response: In3Response):
    """
    Transports each request coming from libin3 to the in3 network and and reports the answer back
    Args:
        in3_request (In3Request): request sent by the In3 Client Core to the In3 Network
        in3_response (In3Response): response to be dispatched to the In3 Client Core
    Returns:
        exit_status (int): Always zero for signaling libin3 the function executed OK.
    """

    for i in range(0, in3_request.urls_len()):
        try:
            endpoint = {
                'host': in3_request.url_at(i),
                'timeout': in3_request.timeout()
            }
            request = {
                'body': in3_request.payload(),
                'headers': {'Content-type': 'application/json', 'Accept': 'application/json'},
            }
            endpoint = http.client.HTTPConnection(**endpoint)
            endpoint.request(**request)
            response = endpoint.getresponse()
            if not response.status == HTTPStatus.OK:
                raise http.client.HTTPException('Request failed with status: {}'.format(str(response.status)))
            if 'error' in response.text:
                in3_response.failure(i, response.content)
            else:
                in3_response.success(i, response.content)
        except http.client.HTTPException as err:
            in3_response.failure(i, str(err).encode('utf8'))
        except Exception as err:
            in3_response.failure(i, str(err).encode('utf8'))
    return 0



"""
Configuration for `in3.client.Client`. Enables mocked in3 network responses to be valid.
Disable node list updates, and collect signed responses from both boot nodes.
"""
from in3 import ClientConfig

_registry = {
    '0x1': {
        'contract': '0xac1b824795e1eb1f6e609fe0da9b9af8beaab60f',
        'registryId': '0x23d5345c5c13180a8080bd5ddbe7cde64683755dcce6e734d95b7b573845facb',
        'needsUpdate': False,
        'avgBlockTime': 15,
        'nodeList': [
            {'url': 'https://in3-v2.slock.it/mainnet/nd-1',
             'address': '0x45d45e6ff99e6c34a235d263965910298985fcfe'},
            {'url': 'https://in3-v2.slock.it/mainnet/nd-2',
             'address': '0x1fe2e9bf29aa1938859af64c413361227d04059a'},
        ]},
    '0x5': {
        'contract': '0x5f51e413581dd76759e9eed51e63d14c8d1379c8',
        'registryId': '0x67c02e5e272f9d6b4a33716614061dd298283f86351079ef903bf0d4410a44ea',
        'needsUpdate': False,
        'avgBlockTime': 15,
        'nodeList': [
            {'url': 'https://in3-v2.slock.it/goerli/nd-1',
             'address': '0x45d45e6ff99e6c34a235d263965910298985fcfe'},
            {'url': 'https://in3-v2.slock.it/goerli/nd-2',
             'address': '0x1fe2e9bf29aa1938859af64c413361227d04059a'},
        ]},
    '0x2a': {
        'contract': '0x4c396dcf50ac396e5fdea18163251699b5fcca25',
        'registryId': '0x92eb6ad5ed9068a24c1c85276cd7eb11eda1e8c50b17fbaffaf3e8396df4becf',
        'needsUpdate': False,
        'avgBlockTime': 6,
        'nodeList': [
            {'url': 'https://in3-v2.slock.it/kovan/nd-1',
             'address': '0x45d45e6ff99e6c34a235d263965910298985fcfe'},
            {'url': 'https://in3-v2.slock.it/kovan/nd-2',
             'address': '0x1fe2e9bf29aa1938859af64c413361227d04059a'},
        ]},
    '0x7d0': {
        'contract': '0xa93b57289070550c82edb1106e12bb37138948b8',
        'registryId': '0xf0162ec6d785ee990e36bad865251f45af0916cf136169540c02b0dd9cb69196',
        'needsUpdate': False,
        'nodeList': [
            {'url': 'https://in3-v2.slock.it/ipfs/nd-1',
             'address': '0x45d45e6ff99e6c34a235d263965910298985fcfe'},
            {'url': 'https://in3-v2.slock.it/ipfs/nd-2',
             'address': '0x1fe2e9bf29aa1938859af64c413361227d04059a'},
        ]},
    '0x99': {
        'contract': '0xed7bb275ca33c46ef3875a9c959c91553ca6acb8',
        'registryId': '0x084ec5cd9274e7c05b827a0d417f92820eb249b9d4ae6e497e355620114a52dc',
        'needsUpdate': False,
        'avgBlockTime': 600,
        'nodeList': [
            {'url': 'https://in3-v2.slock.it/btc/nd-1',
             'address': '0x45d45e6ff99e6c34a235d263965910298985fcfe'},
        ]},
}

mock_config = ClientConfig(node_list_auto_update=False,
                           node_signature_consensus=2,
                           node_signatures=2,
                           latest_block_stall=10,
                           request_retries=1,
                           in3_registry=_registry,
                           cached_blocks=0,
                           cached_code_bytes=0)



import re

from in3.eth.api import EthereumApi
from in3.eth.factory import EthObjectFactory
from in3.libin3.enum import In3Methods
from in3.libin3.runtime import In3Runtime
from in3.model import In3Node, NodeList, ClientConfig, chain_configs
from in3.transport import http_transport


class Client:
    """
    Incubed network client. Connect to the blockchain via a list of bootnodes, then gets the latest list of nodes in
    the network and ask a certain number of the to sign the block header of given list, putting their deposit at stake.
    Once with the latest list at hand, the client can request any other on-chain information using the same scheme.
    Args:
        in3_config (ClientConfig or str): (optional) Configuration for the client. If not provided, default is loaded.
    """

    def __init__(self, chain: str = 'mainnet', in3_config: ClientConfig = None, transport=http_transport):
        if not isinstance(chain, str) or chain.lower() not in ['mainnet', 'kovan', 'goerli']:
            raise AssertionError('Client: Chain name not supported. Try mainnet, kovan, goerli.')
        # TODO: Clear Chain-configs
        if in3_config and not isinstance(in3_config, ClientConfig):
            raise AssertionError('Client: Use in3.ClientConfig to create a new client configuration instance.')

        self._runtime = In3Runtime(chain_configs[chain.lower()].chain_id, transport)
        if in3_config:
            self._configure(in3_config)
        self.eth = EthereumApi(self._runtime)
        self._factory = In3ObjectFactory(self._runtime)

    def _configure(self, in3_config: ClientConfig) -> bool:
        """
        Send RPC to change client configuration. Don't use outside the constructor, might cause instability.
        """
        fn_args = str([in3_config.serialize()]).replace('\'', '')
        return self._runtime.call(In3Methods.CONFIGURE, fn_args, formatted=True)

    def refresh_node_list(self) -> NodeList:
        """
        Gets the list of Incubed nodes registered in the selected chain registry contract.
        Returns:
            node_list (NodeList): List of registered in3 nodes and metadata.
        """
        node_list_dict = self._runtime.call(In3Methods.NODE_LIST)
        return self._factory.get_node_list(node_list_dict)

    def config(self) -> dict:
        # TODO: Marshalling
        return self._runtime.call(In3Methods.GET_CONFIG)

    def ens_namehash(self, domain_name: str) -> str:
        """
        Name format based on [EIP-137](https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax)
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
        Returns:
            node (str): Formatted string referred as `node` in ENS documentation
        """
        if not isinstance(domain_name, str) or not re.match(r'(\w+.eth$)', domain_name):
            raise AssertionError('Client: ENS domain name must end with .ens')
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'hash')

    def ens_address(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to what account that domain points to.
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            address (str): Ethereum address corresponding to what account that domain points to.
        """
        # TODO: Add handlers to Account
        # TODO: Add serialization in account factory
        if registry:
            registry = self._factory.get_address(registry)
        if not isinstance(domain_name, str) or not re.match(r'(\w+.eth$)', domain_name):
            raise AssertionError('Client: ENS domain name must end with .ens')
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'addr', registry)

    def ens_owner(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to Ethereum address of domain owner.
        Args:
            domain_name: ENS supported domain. i.e mydomain.eth
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            owner_address (str): Ethereum address corresponding to domain owner.
        """
        if registry:
            registry = self._factory.get_address(registry)
        if not isinstance(domain_name, str) or not re.match(r'(\w+.eth$)', domain_name):
            raise AssertionError('Client: ENS domain name must end with .ens')
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'owner', registry)

    def ens_resolver(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to Smart-contract address of the resolver registered for that domain.
        Args:
            domain_name: ENS supported domain. i.e mydomain.eth
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            resolver_contract_address (str): Smart-contract address of the resolver registered for that domain.
        """
        if registry:
            registry = self._factory.get_address(registry)
        if not isinstance(domain_name, str) or not re.match(r'(\w+.eth$)', domain_name):
            raise AssertionError('Client: ENS domain name must end with .ens')
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'resolver', registry)


class In3ObjectFactory(EthObjectFactory):

    # TODO: Decode abi for props
    # TODO: Convert registerTime to human readable
    def get_node_list(self, serialized: dict) -> NodeList:
        mapping = {
            "nodes": list,
            "contract": self.get_account,
            "registryId": str,
            "lastBlockNumber": int,
            "totalServers": int,
        }
        deserialized_dict = self._deserialize(serialized, mapping)
        nodes = [self.get_in3node(node) for node in deserialized_dict['nodes']]
        deserialized_dict['nodes'] = nodes
        return NodeList(**deserialized_dict)

    def get_in3node(self, serialized: dict) -> In3Node:
        mapping = {
            "url": str,
            "address": self.get_account,
            "index": int,
            "deposit": self.get_integer,
            "props": str,
            "timeout": int,
            "registerTime": int,
            "weight": int
        }
        deserialized_dict = self._deserialize(serialized, mapping)
        return In3Node(**deserialized_dict)
